/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */


#import "OOSSignature.h"
#import <CommonCrypto/CommonCrypto.h>
#import "OOSCategory.h"
#import "OOSCredentialsProvider.h"
#import "CoreService.h"
#import "OOSTask.h"
#import "OOSCocoaLumberjack.h"

static NSString *const OOSSigV4Marker = @"OOS4";
NSString *const OOSSignatureV4Algorithm = @"OOS";
NSString *const OOSSignatureV4Terminator = @"OOS4_request";

@implementation OOSSignatureSignerUtility

+ (NSData *)sha256HMacWithData:(NSData *)data withKey:(NSData *)key {
    CCHmacContext context;

    CCHmacInit(&context, kCCHmacAlgSHA256, [key bytes], [key length]);
    CCHmacUpdate(&context, [data bytes], [data length]);

    unsigned char digestRaw[CC_SHA256_DIGEST_LENGTH];
    NSInteger digestLength = CC_SHA256_DIGEST_LENGTH;

    CCHmacFinal(&context, digestRaw);

    return [NSData dataWithBytes:digestRaw length:digestLength];
}

+ (NSString *)hashString:(NSString *)stringToHash {
    return [[NSString alloc] initWithData:[self hash:[stringToHash dataUsingEncoding:NSUTF8StringEncoding]]
                                 encoding:NSASCIIStringEncoding];
}

+ (NSData *)hash:(NSData *)dataToHash {
    if ([dataToHash length] > UINT32_MAX) {
        return nil;
    }

    const void *cStr = [dataToHash bytes];
    unsigned char result[CC_SHA256_DIGEST_LENGTH];

    CC_SHA256(cStr, (uint32_t)[dataToHash length], result);

    return [[NSData alloc] initWithBytes:result length:CC_SHA256_DIGEST_LENGTH];
}

+ (NSString *)hexEncode:(NSString *)string {
    NSUInteger len = [string length];
    if (len == 0) {
        return @"";
    }
    unichar *chars = malloc(len * sizeof(unichar));
    if (chars == NULL) {
        // this situation is irrecoverable and we don't want to return something corrupted, so we raise an exception (avoiding NSAssert that may be disabled)
        [NSException raise:@"NSInternalInconsistencyException" format:@"failed malloc" arguments:nil];
        return nil;
    }

    [string getCharacters:chars];

    NSMutableString *hexString = [NSMutableString new];
    for (NSUInteger i = 0; i < len; i++) {
        if ((int)chars[i] < 16) {
            [hexString appendString:@"0"];
        }
        [hexString appendString:[NSString stringWithFormat:@"%x", chars[i]]];
    }
    free(chars);

    return hexString;
}

+ (NSString *)HMACSign:(NSData *)data withKey:(NSString *)key usingAlgorithm:(CCHmacAlgorithm)algorithm {
    CCHmacContext context;
    const char    *keyCString = [key cStringUsingEncoding:NSASCIIStringEncoding];

    CCHmacInit(&context, algorithm, keyCString, strlen(keyCString));
    CCHmacUpdate(&context, [data bytes], [data length]);

    // Both SHA1 and SHA256 will fit in here
    unsigned char digestRaw[CC_SHA256_DIGEST_LENGTH];

    NSInteger digestLength = -1;

    switch (algorithm) {
        case kCCHmacAlgSHA1:
            digestLength = CC_SHA1_DIGEST_LENGTH;
            break;

        case kCCHmacAlgSHA256:
            digestLength = CC_SHA256_DIGEST_LENGTH;
            break;

        default:
            OOSDDLogDebug(@"Unable to sign: unsupported Algorithm.");
            return nil;
            break;
    }

    CCHmacFinal(&context, digestRaw);

    NSData *digestData = [NSData dataWithBytes:digestRaw length:digestLength];

    return [digestData base64EncodedStringWithOptions:kNilOptions];
}

@end

#pragma mark - OOSSignatureV4Signer

@interface OOSSignatureV4Signer()

@property (nonatomic, strong) OOSEndpoint *endpoint;

@end

@implementation OOSSignatureV4Signer

+ (instancetype)signerWithCredentialsProvider:(id<OOSCredentialsProvider>)credentialsProvider
                                     endpoint:(OOSEndpoint *)endpoint {
    OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:credentialsProvider
                                                                                    endpoint:endpoint];
    return signer;
}

- (instancetype)initWithCredentialsProvider:(id<OOSCredentialsProvider>)credentialsProvider
                                   endpoint:(OOSEndpoint *)endpoint {
    if (self = [super init]) {
        _credentialsProvider = credentialsProvider;
        _endpoint = endpoint;
    }

    return self;
}

- (OOSTask *)interceptRequest:(NSMutableURLRequest *)request {
    [request addValue:request.URL.host forHTTPHeaderField:@"Host"];
	
    return [[self.credentialsProvider credentials] continueWithSuccessBlock:^id _Nullable(OOSTask<OOSCredentials *> * _Nonnull task) {
        OOSCredentials *credentials = task.result;
        // clear authorization header if set
        [request setValue:nil forHTTPHeaderField:@"Authorization"];

        if (credentials) {
			NSString *authorization = [self signS3RequestV4:request credentials:credentials];

            if (authorization) {
                [request setValue:authorization forHTTPHeaderField:@"Authorization"];
            }
        }
        return nil;
    }];
}

- (NSString *)signS3RequestV4:(NSMutableURLRequest *)urlRequest
                  credentials:(OOSCredentials *)credentials {
	
	NSMutableDictionary *headers = [[urlRequest allHTTPHeaderFields] mutableCopy];

	NSString *contentType = [urlRequest valueForHTTPHeaderField:@"Content-Type"];
	if (contentType == nil) {
		contentType = @"";
	}

    NSDate *date = [NSDate OOS_clockSkewFixedDate];
	
	NSString *dateTime  = [date OOS_stringValue:OOSDateRFC822DateFormat1];
	[urlRequest setValue:dateTime forHTTPHeaderField:@"Date"];
	
    // compute canonical request
    NSString *httpMethod = urlRequest.HTTPMethod;
    // URL.path returns unescaped path
    // For S3,  url-encoded URI need to be decoded before generate  CanonicalURI, otherwise, signature doesn't match occurs. (I.e. CanonicalURI for "/ios-v2-test-445901470/name%3A" will still be  "/ios-v2-test-445901470/name%3A".  "%3A" -> ":" -> "%3A")
    NSString *cfPath = (NSString*)CFBridgingRelease(CFURLCopyPath((CFURLRef)urlRequest.URL));
    NSString *path = [cfPath OOS_stringWithURLEncodingPath];
    
    if (path.length == 0) {
        path = [NSString stringWithFormat:@"/"];
    }
	
	NSString *query = urlRequest.URL.query;
    if (query == nil) {
        query = [NSString stringWithFormat:@""];
    }

    // Compute contentSha256
    NSUInteger contentLength = [[urlRequest allHTTPHeaderFields][@"Content-Length"] integerValue];
	
	//using Content-Length with value of '0' cause auth issue, remove it.
	if (contentLength == 0) {
		[urlRequest setValue:nil forHTTPHeaderField:@"Content-Length"];
	}

    //Set Content-MD5 header field if required by server.
    if (([ urlRequest.HTTPMethod isEqualToString:@"PUT"] && ([[[urlRequest URL] query] hasPrefix:@"tagging"] ||
                                                             [[[urlRequest URL] query] hasPrefix:@"lifecycle"] ||
                                                             [[[urlRequest URL] query] hasPrefix:@"cors"]))
        || ([urlRequest.HTTPMethod isEqualToString:@"POST"] && [[[urlRequest URL] query] hasPrefix:@"delete"])
		|| ([urlRequest.HTTPMethod isEqualToString:@"POST"] && [[[urlRequest URL] query] hasPrefix:@"uploadId"])
        ) {
        if (![urlRequest valueForHTTPHeaderField:@"Content-MD5"]) {
            [urlRequest setValue:[NSString OOS_base64md5FromData:urlRequest.HTTPBody] forHTTPHeaderField:@"Content-MD5"];
        }
    }

	//	StringToSign = HTTP-VERB + "\n" +
	//	Content-MD5 + "\n" +
	//	Content-Type + "\n" +
	//	Date + "\n" +
	//	CanonicalizedAmzHeaders +
	//	CanonicalizedResource;
	
	NSString *contentMD5 = [urlRequest valueForHTTPHeaderField:@"Content-MD5"];
	if (contentMD5 == nil) {
		contentMD5 = @"";
	}
	NSString *CanonicalizedAmzHeaders = [OOSSignatureV4Signer getCanonicalizedHeaderString:headers];
	NSString *CanonicalizedResource = [OOSSignatureV4Signer getCanonicalizedResource:httpMethod path:path query:query headers:headers];

	NSString *dateString = [urlRequest valueForHTTPHeaderField:@"Date"];
	if (dateString == nil) {
		dateString = @"";
	}
    NSString *stringToSign = [NSString stringWithFormat:@"%@\n%@\n%@\n%@\n%@%@",
                              urlRequest.HTTPMethod,
							  contentMD5,
							  contentType,
                              dateString,
                              CanonicalizedAmzHeaders,
                              CanonicalizedResource];
    OOSDDLogDebug(@"String to Sign: %@", stringToSign);
	
	NSData *stringToSignData = [stringToSign dataUsingEncoding:NSUTF8StringEncoding];

	NSString *signatureString = [OOSSignatureSignerUtility HMACSign:stringToSignData withKey:credentials.secretKey usingAlgorithm:kCCHmacAlgSHA1];

	NSString *prefix = @"OOS";
	// 批量删除接口需要设置为AWS
	if ([query isEqualToString:@"delete="]) {
		prefix = @"AWS";
	}
	
	NSString *accessKey = credentials.accessKey;
	NSString *authorization = [NSString stringWithFormat:@"%@ %@:%@",
							   prefix,
                               accessKey,
                               signatureString];

    return authorization;
}


+ (OOSTask<NSURL *> *)generateQueryStringForSignatureV4WithCredentialProvider:(id<OOSCredentialsProvider>)credentialsProvider
																   httpMethod:(OOSHTTPMethod)httpMethod
															   expireDuration:(int32_t)expireDuration
																	 endpoint:(OOSEndpoint *)endpoint
																	  keyPath:(NSString *)keyPath
															   requestHeaders:(NSDictionary<NSString *, NSString *> *)requestHeaders
															requestParameters:(NSDictionary<NSString *, id> *)requestParameters
																	 signBody:(BOOL)signBody{
	
	return [[credentialsProvider credentials] continueWithSuccessBlock:^id _Nullable(OOSTask<OOSCredentials *> * _Nonnull task) {
		OOSCredentials *credentials = task.result;
		
		//Implementation of V4 signaure http://docs.OOS.amazon.com/AmazonS3/latest/API/sigv4-query-string-auth.html
		NSMutableString *queryString = [NSMutableString new];
		
		//Append Identifies the version of OOS Signature and the algorithm that you used to calculate the signature.
//		[queryString appendFormat:@"%@=%@&",@"X-Amz-Algorithm",OOSSignatureV4Algorithm];
		
		//Get ClockSkew Fixed Date
		NSDate *currentDate = [NSDate OOS_clockSkewFixedDate];
		
		//Format of X-Amz-Credential : <your-access-key-id>/<date>/<OOS-region>/.
		NSString *scope = [NSString stringWithFormat:@"%@/%@/",
						   [currentDate OOS_stringValue:OOSDateShortDateFormat1],
						   endpoint.regionName];
		
		NSString *signingCredentials = [NSString stringWithFormat:@"%@/%@",credentials.accessKey, scope];
		//need to replace "/" with "%2F"
		NSString *xAmzCredentialString = [signingCredentials stringByReplacingOccurrencesOfString:@"/" withString:@"\%2F"];
		
		[queryString appendFormat:@"%@=%@&",@"X-Amz-Credential",xAmzCredentialString];
		
		//X-Amz-Date in ISO 8601 format, for example, 20130721T201207Z. This value must match the date value used to calculate the signature.
		[queryString appendFormat:@"%@=%@&",@"X-Amz-Date",[currentDate OOS_stringValue:OOSDateISO8601DateFormat2]];
		
		//X-Amz-Expires, Provides the time period, in seconds, for which the generated presigned URL is valid.
		//For example, 86400 (24 hours). This value is an integer. The minimum value you can set is 1, and the maximum is 604800 (seven days).
		[queryString appendFormat:@"%@=%d&", @"X-Amz-Expires", expireDuration];
		
		/*
		 X-Amz-SignedHeaders Lists the headers that you used to calculate the signature. The HTTP host header is required.
		 Any x-amz-* headers that you plan to add to the request are also required for signature calculation.
		 In general, for added security, you should sign all the request headers that you plan to include in your request.
		 */
		
//		[queryString appendFormat:@"%@=%@&", @"X-Amz-SignedHeaders", [[OOSSignatureV4Signer getSignedHeadersString:requestHeaders] OOS_stringWithURLEncoding]];
		
		//add additionalParameters to queryString
		for (NSString *key in requestParameters) {
			if ([requestParameters[key] isKindOfClass:[NSArray class]]) {
				NSArray<NSString *> *parameterValues = requestParameters[key];
				for (NSString *paramValue in parameterValues) {
					[queryString appendFormat:@"%@=%@&", [key OOS_stringWithURLEncoding], [paramValue OOS_stringWithURLEncoding]];
				}
			} else if ([requestParameters[key] isKindOfClass:[NSString class]]) {
				NSString *value = requestParameters[key];
				[queryString appendFormat:@"%@=%@&",[key OOS_stringWithURLEncoding], [value OOS_stringWithURLEncoding]];
			} else {
				// Only @[NSString: NSString] and @[NSString: NSArray<NSString>] supported currently
				@throw [NSException exceptionWithName:NSInternalInconsistencyException
											   reason:@"Invalid requestParameters dictionary. Supported Dictionaries include [NSString: NSString] and [NSString: NSArray<NSString>]"
											 userInfo:nil];
			}
		}
		
		// =============  generate v4 signature string ===================
		
		/* Canonical Request Format:
		 *
		 * HTTP-VERB + "\n" +  (e.g. GET, PUT, POST)
		 * Canonical URI + "\n" + (e.g. /test.txt)
		 * Canonical Query String + "\n" (multiple queryString need to sorted by QueryParameter)
		 * Canonical Headrs + "\n" + (multiple headers need to be sorted by HeaderName)
		 * Signed Headers + "\n" + (multiple headers need to be sorted by HeaderName)
		 * "UNSIGNED-PAYLOAD"
		 */
		
		
		NSString *httpMethodString = [NSString oos_stringWithHTTPMethod:httpMethod];
		
		//CanonicalURI is the URI-encoded version of the absolute path component of the URI—everything starting with the "/" that follows the domain name and up to the end of the string or to the question mark character ('?') if you have query string parameters. e.g. https://s3.amazonOOS.com/examplebucket/myphoto.jpg /examplebucket/myphoto.jpg is the absolute path. In the absolute path, you don't encode the "/".
		
		NSString *canonicalURI = [NSString stringWithFormat:@"/%@", [keyPath OOS_stringWithURLEncodingPath]]; //keyPath is not url-encoded.
		
		NSString *contentSha256;
		if(signBody && httpMethod == OOSHTTPMethodGET){
			//in case of http get we sign the body as an empty string only if the sign body flag is set to true
			contentSha256 = [OOSSignatureSignerUtility hexEncode:[[NSString alloc] initWithData:[OOSSignatureSignerUtility hash:[@"" dataUsingEncoding:NSUTF8StringEncoding]] encoding:NSASCIIStringEncoding]];
		}else{
			contentSha256 = @"UNSIGNED-PAYLOAD";
		}
		//Generate Canonical Request
		NSString *canonicalRequest = [OOSSignatureV4Signer getCanonicalizedRequest:httpMethodString
																			  path:canonicalURI
																			 query:queryString
																		   headers:requestHeaders
																	 contentSha256:contentSha256];
		
		//Generate String to Sign
		NSString *stringToSign = [NSString stringWithFormat:@"%@\n%@\n%@\n%@",
								  OOSSignatureV4Algorithm,
								  [currentDate OOS_stringValue:OOSDateISO8601DateFormat2],
								  scope,
								  [OOSSignatureSignerUtility hexEncode:[OOSSignatureSignerUtility hashString:canonicalRequest]]];
		
		//Generate Signature
		NSData *kSigning  = [OOSSignatureV4Signer getV4DerivedKey:credentials.secretKey
															 date:[currentDate OOS_stringValue:OOSDateShortDateFormat1]
														   region:endpoint.regionName];
		
		NSData *signature = [OOSSignatureSignerUtility sha256HMacWithData:[stringToSign dataUsingEncoding:NSUTF8StringEncoding]
																  withKey:kSigning];
		NSString *signatureString = [OOSSignatureSignerUtility hexEncode:[[NSString alloc] initWithData:signature
																							   encoding:NSASCIIStringEncoding]];
		
		// ============  generate v4 signature string (END) ===================
		
		[queryString appendFormat:@"%@=%@", @"X-Amz-Signature", signatureString];
		
		NSString *urlString = [NSString stringWithFormat:@"%@://%@/%@?%@", endpoint.URL.scheme, endpoint.hostName, keyPath, queryString];
		
		return [NSURL URLWithString:urlString];
	}];
}

+ (NSString *)getCanonicalizedResource:(NSString *)method path:(NSString *)path query:(NSString *)query headers:(NSDictionary *)headers {
	NSMutableString *canonicalResource = [NSMutableString new];
	[canonicalResource appendString:path];
	
	// add bucket
	NSString *host = headers[@"host"];
	NSArray *hostItems = [host componentsSeparatedByString:@"."];
	if ([hostItems count] >= 4) {
		[canonicalResource insertString:[NSString stringWithFormat:@"/%@", [hostItems firstObject]] atIndex:0];
	}
	
	// add sub resource
	const NSArray *resourcesArray = @[ @"acl", @"lifecycle", @"location", @"logging", @"notification", @"partNumber", @"policy", @"requestPayment", @"torrent", @"uploadId", @"uploads", @"versionId", @"versioning", @"versions", @"website", @"delete", @"cors" ];
	
	__block NSMutableArray *subResourceArray = [NSMutableArray new];
	[[query componentsSeparatedByString:@"&"] enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
		NSArray *components = [obj componentsSeparatedByString:@"="];
		NSString *key;
		NSUInteger count = [components count];
		if (count > 0 && count <= 2) {
			//can be ?a=b or ?a
			key = components[0];
			if  (! [key isEqualToString:@""] && [resourcesArray indexOfObject:key] != NSNotFound ) {
				[subResourceArray addObject:key];
			}
		}
	}];
	
	[subResourceArray sortUsingSelector:@selector(compare:)];
	NSString *subResource = [subResourceArray componentsJoinedByString:@"&"];
	
	if (subResource != nil && subResource.length > 0 ) {
		if ([method isEqualToString:@"DELETE"] && [subResource isEqualToString:@"trigger"]) {
			[canonicalResource appendFormat:@"?%@", query];
		} else if ([subResource containsString:@"uploadId"] ) {
			[canonicalResource appendFormat:@"?%@", query];
		} else {
			[canonicalResource appendFormat:@"?%@", subResource];
		}
	}
	
	return canonicalResource;
}

+ (NSString *)getCanonicalizedRequest:(NSString *)method path:(NSString *)path query:(NSString *)query headers:(NSDictionary *)headers contentSha256:(NSString *)contentSha256 {
    NSMutableString *canonicalRequest = [NSMutableString new];
    [canonicalRequest appendString:method];
    [canonicalRequest appendString:@"\n"];
    [canonicalRequest appendString:path]; // Canonicalized resource path
    [canonicalRequest appendString:@"\n"];

    [canonicalRequest appendString:[OOSSignatureV4Signer getCanonicalizedQueryString:query]]; // Canonicalized Query String
    [canonicalRequest appendString:@"\n"];

    [canonicalRequest appendString:[OOSSignatureV4Signer getCanonicalizedHeaderString:headers]];
    [canonicalRequest appendString:@"\n"];

    [canonicalRequest appendString:[OOSSignatureV4Signer getSignedHeadersString:headers]];
    [canonicalRequest appendString:@"\n"];

    [canonicalRequest appendString:[NSString stringWithFormat:@"%@", contentSha256]];

    return canonicalRequest;
}

+ (NSString *)getCanonicalizedQueryString:(NSString *)query {
    NSMutableDictionary<NSString *, NSMutableArray<NSString *> *> *queryDictionary = [NSMutableDictionary new];
    [[query componentsSeparatedByString:@"&"] enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
        NSArray *components = [obj componentsSeparatedByString:@"="];
        NSString *key;
        NSString *value = @"";
        NSUInteger count = [components count];
        if (count > 0 && count <= 2) {
            //can be ?a=b or ?a
            key = components[0];
            if  (! [key isEqualToString:@""] ) {
                if (count == 2) {
                    //is ?a=b
                    value = components[1];
                }
                if (queryDictionary[key]) {
                    // If the query parameter has multiple values, add it in the mutable array
                    [[queryDictionary objectForKey:key] addObject:value];
                } else {
                    // Insert the value for query parameter as an element in mutable array
                    [queryDictionary setObject:[@[value] mutableCopy] forKey:key];
                }
            }
        }
    }];

    NSMutableArray *sortedQuery = [[NSMutableArray alloc] initWithArray:[queryDictionary allKeys]];

    [sortedQuery sortUsingSelector:@selector(compare:)];

    NSMutableString *sortedQueryString = [NSMutableString new];
    for (NSString *key in sortedQuery) {
        [queryDictionary[key] sortUsingSelector:@selector(compare:)];
        for (NSString *parameterValue in queryDictionary[key]) {
            [sortedQueryString appendString:key];
            [sortedQueryString appendString:@"="];
            [sortedQueryString appendString:parameterValue];
            [sortedQueryString appendString:@"&"];
        }
    }
    // Remove the trailing & for a valid canonical query string.
    if ([sortedQueryString hasSuffix:@"&"]) {
        return [sortedQueryString substringToIndex:[sortedQueryString length] - 1];
    }

    return sortedQueryString;
}

+ (NSString *)getCanonicalizedHeaderString:(NSDictionary *)headers {
    NSMutableArray *sortedHeaders = [[NSMutableArray alloc] initWithArray:[headers allKeys]];

    [sortedHeaders sortUsingSelector:@selector(caseInsensitiveCompare:)];

    NSMutableString *headerString = [NSMutableString new];
    for (NSString *header in sortedHeaders) {
		NSString *lowerHeader = [header lowercaseString];
		if ( [lowerHeader hasPrefix:@"x-amz-"] ) {
			[headerString appendString:[header lowercaseString]];
			[headerString appendString:@":"];
			[headerString appendString:[headers valueForKey:header]];
			[headerString appendString:@"\n"];
		}
    }

    // SigV4 expects all whitespace in headers and values to be collapsed to a single space
    NSCharacterSet *whitespaceChars = [NSCharacterSet whitespaceCharacterSet];
    NSPredicate *noEmptyStrings = [NSPredicate predicateWithFormat:@"SELF != ''"];

    NSArray *parts = [headerString componentsSeparatedByCharactersInSet:whitespaceChars];
    NSArray *nonWhitespace = [parts filteredArrayUsingPredicate:noEmptyStrings];
    return [nonWhitespace componentsJoinedByString:@" "];
}

+ (NSString *)getSignedHeadersString:(NSDictionary *)headers {
    NSMutableArray *sortedHeaders = [[NSMutableArray alloc] initWithArray:[headers allKeys]];

    [sortedHeaders sortUsingSelector:@selector(caseInsensitiveCompare:)];

    NSMutableString *headerString = [NSMutableString new];
    for (NSString *header in sortedHeaders) {
        if ([headerString length] > 0) {
            [headerString appendString:@";"];
        }
        [headerString appendString:[header lowercaseString]];
    }

    return headerString;
}

+ (NSData *)getV4DerivedKey:(NSString *)secret date:(NSString *)dateStamp region:(NSString *)regionName {
    // OOS4 uses a series of derived keys, formed by hashing different pieces of data
    NSString *kSecret = [NSString stringWithFormat:@"%@%@", OOSSigV4Marker, secret];
    NSData *kDate = [OOSSignatureSignerUtility sha256HMacWithData:[dateStamp dataUsingEncoding:NSUTF8StringEncoding]
                                                          withKey:[kSecret dataUsingEncoding:NSUTF8StringEncoding]];
    NSData *kRegion = [OOSSignatureSignerUtility sha256HMacWithData:[regionName dataUsingEncoding:NSASCIIStringEncoding]
                                                            withKey:kDate];
    NSData *kSigning = [OOSSignatureSignerUtility sha256HMacWithData:[OOSSignatureV4Terminator dataUsingEncoding:NSUTF8StringEncoding]
                                                             withKey:kRegion];

    //TODO: cache this derived key?
    return kSigning;
}

@end

