//
//  OOSNetworkingTests.m
//  OOSTests
//
//  Created by NorrisTong on 2018/10/27.
//  Copyright © 2018 CTYun. All rights reserved.
//

#import <XCTest/XCTest.h>
#import "OOSNetworking.h"
#import "OOSTask.h"
#import "OOSCancellationToken.h"
#import "OOSCancellationTokenSource.h"
#import "OOSURLResponseSerialization.h"
#import "OOSSignature.h"
#import "OOSCredentialsProvider.h"
#import "CoreService.h"
#import "OOSResources.h"

@interface OOSNetworkingTests : XCTestCase

@end

@implementation OOSNetworkingTests

NSString *const AccessKey = @"2510935f75d781385b35";
NSString *const SecretKey = @"567b89b544a89bb3d3ccef4be8507e6d30d75579";

- (void)setUp {
    // Put setup code here. This method is called before the invocation of each test method in the class.
}

- (void)tearDown {
    // Put teardown code here. This method is called after the invocation of each test method in the class.
}

- (void)testCancelSimple {
	OOSCancellationTokenSource *cts = [OOSCancellationTokenSource cancellationTokenSource];
	OOSTask *task = [OOSTask taskWithDelay:100];
	
	task = [task continueWithSuccessBlock:^id(OOSTask *t) {
		XCTFail(@"Success block should not be triggered");
		return nil;
	} cancellationToken:cts.token];
	
	[cts cancel];
	[task waitUntilFinished];
	
	XCTAssertTrue(task.cancelled);
}

- (void)testCancelBlock {
    XCTestExpectation* expectation = [self expectationWithDescription:@"testCancelBlock"];
	
    OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
    OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
    OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];

    OOSTask *task = [networking sendRequest:networkingRequest];
	
	OOSCancellationTokenSource *source = [OOSCancellationTokenSource cancellationTokenSource];
    OOSCancellationToken *cancelToken = source.token;
    [cancelToken registerCancellationObserverWithBlock:^{
        [expectation fulfill];
    }];
	
    [task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
        XCTFail(@"Success block should not be triggered");
        return t;
    } cancellationToken:cancelToken];

	[source cancel];
	[task waitUntilFinished];

    [self waitForExpectationsWithTimeout:4 handler:nil];
}

- (void)testCancelContinue {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testCancelContinue"];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = [NSURL URLWithString:@"https://www.google.com"];
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	
	OOSCancellationTokenSource *source = [OOSCancellationTokenSource cancellationTokenSource];
	OOSCancellationToken *cancelToken = source.token;
	[cancelToken registerCancellationObserverWithBlock:^{
		XCTFail(@"cancel block should not be triggered");
	}];
	
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull task) {
		XCTAssertNotNil(task.error);
		XCTAssertEqual(OOSNetworkingErrorCancelled, task.error.code);
		
		[expectation fulfill];
		return task;
	} cancellationToken:cancelToken];
	
	dispatch_after(dispatch_time(DISPATCH_TIME_NOW, 0.1 * NSEC_PER_SEC), dispatch_get_main_queue(), ^{
		[networkingRequest cancel];
	});
	
	[self waitForExpectationsWithTimeout:4 handler:nil];
}

- (void)testPause {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testPause"];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = [NSURL URLWithString:@"https://www.google.com"];
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	
	OOSCancellationTokenSource *source = [OOSCancellationTokenSource cancellationTokenSource];
	OOSCancellationToken *cancelToken = source.token;
	[cancelToken registerCancellationObserverWithBlock:^{
		XCTFail(@"cancel block should not be triggered");
	}];
	
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNotNil(task.error);
		XCTAssertEqual(OOSNetworkingErrorCancelled, task.error.code);
		
		[expectation fulfill];
		return t;
	} cancellationToken:cancelToken];
	
	dispatch_after(dispatch_time(DISPATCH_TIME_NOW, 0.1 * NSEC_PER_SEC), dispatch_get_main_queue(), ^{
		[networkingRequest pause];
	});
	
	[self waitForExpectationsWithTimeout:4 handler:nil];
}

- (void)testSuccessBlock {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSuccessBlock"];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = [NSURL URLWithString:@"https://oos.ctyunapi.cn"];
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodGET;
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	
	[task continueWithSuccessBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertFalse(task.isCancelled);
		XCTAssertNil(task.error);
		
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:4 handler:nil];
}

- (void)testXMLResponseSerialization {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testXMLResponseSerialization"];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = [NSURL URLWithString:@"https://oos.ctyunapi.cn"];
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:nil
																					  actionName:@"operationName"
																					 outputClass:[NSString class]];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodGET;
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);

		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:4 handler:nil];
}

- (void)testSignatureListBuckets {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSignatureListBuckets"];
	
	OOSEndpoint *endpoint = [[OOSEndpoint alloc] initWithRegion:OOSRegionHangZhou];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = endpoint.URL;
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:[[OOSResources sharedInstance] JSONObject]
																					 actionName:@"ListBuckets"
																					outputClass:nil];
	
	OOSStaticCredentialsProvider *provider = [[OOSStaticCredentialsProvider alloc] initWithAccessKey:AccessKey secretKey:SecretKey];
	
	OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:provider endpoint:endpoint];
	configuration.requestInterceptors = @[signer];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodGET;
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:10 handler:nil];
}

- (void)testSignatureListObjects {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSignatureListObjects"];
	
	OOSEndpoint *endpoint = [[OOSEndpoint alloc] initWithRegion:OOSRegionHangZhou];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = endpoint.URL;
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:[[OOSResources sharedInstance] JSONObject]
																					 actionName:@"ListObjects"
																					outputClass:nil];
	
	OOSStaticCredentialsProvider *provider = [[OOSStaticCredentialsProvider alloc] initWithAccessKey:AccessKey secretKey:SecretKey];
	
	OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:provider endpoint:endpoint];
	configuration.requestInterceptors = @[signer];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodGET;
	networkingRequest.URLString = @"https://norris.oos-hz.ctyunapi.cn/?prefix=onepiece&max-keys=50";
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:10 handler:nil];
}

- (void)testSignatureGetObject {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSignatureGetObject"];
	
	OOSEndpoint *endpoint = [[OOSEndpoint alloc] initWithRegion:OOSRegionHangZhou];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = endpoint.URL;
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:[[OOSResources sharedInstance] JSONObject]
																					 actionName:@"GetObject"
																					outputClass:nil];
	
	OOSStaticCredentialsProvider *provider = [[OOSStaticCredentialsProvider alloc] initWithAccessKey:AccessKey secretKey:SecretKey];
	
	OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:provider endpoint:endpoint];
	configuration.requestInterceptors = @[signer];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodGET;
	networkingRequest.URLString = @"https://norris.oos-hz.ctyunapi.cn/photos/puppy.jpg";
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:10 handler:nil];
}

- (void)testSignatureDeleteObject {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSignatureDeleteObject"];
	
	OOSEndpoint *endpoint = [[OOSEndpoint alloc] initWithRegion:OOSRegionHangZhou];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = endpoint.URL;
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:[[OOSResources sharedInstance] JSONObject]
																					 actionName:@"DeleteObject"
																					outputClass:nil];
	
	OOSStaticCredentialsProvider *provider = [[OOSStaticCredentialsProvider alloc] initWithAccessKey:AccessKey secretKey:SecretKey];
	
	OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:provider endpoint:endpoint];
	configuration.requestInterceptors = @[signer];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodDELETE;
	networkingRequest.URLString = @"https://norris.oos-hz.ctyunapi.cn/d.jpg";
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:10 handler:nil];
}

- (void)testSignatureCreateBucket {
	XCTestExpectation* expectation = [self expectationWithDescription:@"testSignatureCreateBucket"];
	
	OOSEndpoint *endpoint = [[OOSEndpoint alloc] initWithRegion:OOSRegionHangZhou];
	
	OOSNetworkingConfiguration *configuration = [OOSNetworkingConfiguration new];
	configuration.baseURL = endpoint.URL;
	configuration.responseSerializer = [[OOSXMLResponseSerializer alloc] initWithJSONDefinition:[[OOSResources sharedInstance] JSONObject]
																					 actionName:@"CreateBucket"
																					outputClass:nil];
	
	OOSStaticCredentialsProvider *provider = [[OOSStaticCredentialsProvider alloc] initWithAccessKey:AccessKey secretKey:SecretKey];
	
	OOSSignatureV4Signer *signer = [[OOSSignatureV4Signer alloc] initWithCredentialsProvider:provider endpoint:endpoint];
	configuration.requestInterceptors = @[signer];
	
	OOSNetworking *networking = [[OOSNetworking alloc] initWithConfiguration:configuration];
	OOSNetworkingRequest *networkingRequest = [OOSNetworkingRequest new];
	networkingRequest.HTTPMethod = OOSHTTPMethodPUT;
	
	NSTimeInterval timeIntervalSinceReferenceDate = [NSDate timeIntervalSinceReferenceDate];
	networkingRequest.URLString = [NSString stringWithFormat:@"https://norris-%lld.oos-hz.ctyunapi.cn/", (int64_t)timeIntervalSinceReferenceDate];
	
	OOSTask *task = [networking sendRequest:networkingRequest];
	[task continueWithBlock:^id _Nullable(OOSTask * _Nonnull t) {
		XCTAssertNil(t.error);
		[expectation fulfill];
		return t;
	}];
	
	[self waitForExpectationsWithTimeout:10 handler:nil];
}

@end
